/* /////////////////////////////////////////////////////////////////////////
 * File:        src/xtests.core.cpp (formerly part of Synesis' internal test codebase)
 *
 * Purpose:     Implementation for xTests core library.
 *
 * Created:     20th June 1999
 * Updated:     23rd February 2008
 *
 * Home:        http://stlsoft.org/
 *
 * Copyright (c) 1999-2008, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the
 *   names of any contributors may be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////// */


/* xTests Header Files */
#define _XTESTS_NO_CPP_API
#include <xtests/xtests.h>
#include <xtests/internal/safestr.h>

/* Warning suppressions */
#if defined(STLSOFT_COMPILER_IS_MSVC)
# ifdef XTESTS_USING_SAFE_STR_FUNCTIONS
#  pragma warning(disable : 4996)
# endif /* XTESTS_USING_SAFE_STR_FUNCTIONS */
# if defined(STLSOFT_CF_EXCEPTION_SUPPORT)
#  pragma warning(disable : 4702)
# else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
#  pragma warning(disable : 4530)
# endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
#endif /* compiler */

#if defined(XTESTS_USE_PANTHEIOS)
# include <pantheios/pantheios.h>
#endif /* XTESTS_USE_PANTHEIOS */

/* STLSoft Header Files */
#include <stlsoft/memory/auto_buffer.hpp>
#include <stlsoft/shims/access/string/std/c_string.h>
#include <stlsoft/string/case_functions.hpp>
#include <stlsoft/string/string_traits_fwd.hpp>
#include <stlsoft/util/must_init.hpp>
#include <stlsoft/util/integral_printf_traits.hpp>

/* Standard C++ Header Files */
#include <map>
#include <string>

/* Standard C Header Files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Warning suppressions */
#if defined(STLSOFT_COMPILER_IS_MSVC)
# if defined(STLSOFT_CF_EXCEPTION_SUPPORT)
# else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
#  pragma warning(default : 4530)
# endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////
 * Compatiblity tests
 */

#ifndef STLSOFT_CF_RTTI_SUPPORT
# error Can only be used when RTTI-support is enabled
#endif /* !STLSOFT_CF_RTTI_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

namespace
{

    typedef stlsoft::auto_buffer<char>  char_buffer_t_;

} // anonymous namespace

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

    STLSOFT_TEMPLATE_SPECIALISATION
    struct string_traits< char_buffer_t_>
    {
        typedef char_buffer_t_::value_type  char_type;
    };

#ifndef _STLSOFT_NO_NAMESPACE
} /* namespace stlsoft */
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _XTESTS_NO_NAMESPACE
namespace xtests
{
namespace c
{
#endif /* !_XTESTS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
namespace
{
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
    typedef stlsoft::ss_sint64_t                sint64_t;
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */

    typedef std::string                         string_t;

    struct TestInfo
    {
        const string_t  name;
        const string_t  description;
        unsigned        totalConditions;
        unsigned        failedConditions;
        unsigned        unexpectedExceptions;
        unsigned        expectedExceptions;

        TestInfo(char const* name, char const* description)
            : name(name)
            , description(description)
            , totalConditions(0)
            , failedConditions(0)
            , unexpectedExceptions(0)
            , expectedExceptions(0)
        {}

        bool haveAllTestsPassed() const
        {
            return (0 == failedConditions && 0 ==  unexpectedExceptions && 0 == expectedExceptions);
        }

    private:
        TestInfo &operator =(TestInfo const&);
    };


    typedef std::map<string_t, TestInfo>        test_map_t;

    class RunnerInfo
    {
    public:
        RunnerInfo(char const* name, int verbosity, xTests_Reporter_t* reporter, void* reporterParam);

        int BeginCase(char const* name, char const* description);
        int EndCase(char const* name);

        int RegisterSuccessfulCondition();
        int RegisterFailedCondition(   char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   char const*         expr);
        int RegisterFailedCondition_long(char const* file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   long                expected
                                    ,   long                actual
                                    ,   xtests_comparison_t comp);
        int RegisterFailedCondition_ulong(char const* file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   unsigned long       expected
                                    ,   unsigned long       actual
                                    ,   xtests_comparison_t comp);
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
        int RegisterFailedCondition_longlong(char const*   file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   sint64_t            expected
                                    ,   sint64_t            actual
                                    ,   xtests_comparison_t comp);
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
        int RegisterFailedCondition_double(char const*     file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   double              expected
                                    ,   double              actual
                                    ,   xtests_comparison_t comp);
        int TestMultibyteStrings(       char const*     file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   char const*         expected
                                    ,   char const*         actual
                                    ,   xtests_comparison_t comp);
        int WriteFailMessage(           char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   char const*         message
                                    ,   char const*         qualifyingInformation);
        void CaseExcepted(  char const* exceptionType
                        ,   char const* exceptionMessage);
        void CaseExceptionExpected(char const* exceptionType);

        void PrintStart();
        void PrintResults();
        void onAbend(char const* message);
        int NumberOfFailedTestCases() const;

    private:
        static xTests_Reporter_t*   get_reporter_(xTests_Reporter_t*);

    private:
#if defined(STLSOFT_INCL_STLSOFT_UTIL_HPP_MUST_INIT)
        typedef stlsoft::must_init<int>         int_type;
        typedef stlsoft::must_init<unsigned>    unsigned_type;
#else /* ? STLSOFT_INCL_STLSOFT_UTIL_HPP_MUST_INIT */
        typedef int                             int_type;
        typedef unsigned                        unsigned_type;
#endif /* STLSOFT_INCL_STLSOFT_UTIL_HPP_MUST_INIT */

        xTests_Reporter_t* const    m_reporter;
        void* const                 m_reporterParam;
        const string_t              m_name;
        const int_type              m_verbosity;
        test_map_t                  m_tests;
        unsigned_type               m_totalConditions;
        unsigned_type               m_failedConditions;
        unsigned_type               m_unexpectedExceptions;
        unsigned_type               m_expectedExceptions;
        test_map_t::iterator        m_currentCase;

    private:
        RunnerInfo &operator =(RunnerInfo const&);
    };

    ////////////////////////////////////////////////////////////////////////
    // Non-local variables

    static RunnerInfo*      s_runner;
    static double           s_fpApproximateFactor   =   XTESTS_FP_APPROXIMATE_FACTOR;

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
} // anonymous namespace
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////
 * Helper functions
 */

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
namespace
{
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

    enum xtests_severity_t
    {
            xtestsEmergency     =   0
        ,   xtestsAlert         =   1
        ,   xtestsCritical      =   2
        ,   xtestsError         =   3
        ,   xtestsWarning       =   4
        ,   xtestsNotice        =   5
        ,   xtestsInformational =   6
        ,   xtestsDebug         =   7
    };

    static int xtests_output_(xtests_severity_t sev, char const* message)
    {
#if defined(XTESTS_USE_PANTHEIOS)

        pantheios::puts(sev, message);

        return 1;

#else /* ? XTESTS_USE_PANTHEIOS */

        FILE*   stm = (sev < xtestsNotice) ? stderr : stdout;

        return ::fprintf(stm, message);

#endif /* XTESTS_USE_PANTHEIOS */
    }

    static int xtests_strcmp_(char const* str1, char const* str2)
    {
        if(str1 == str2)
        {
            return 0;
        }
        else if(NULL == str1)
        {
            STLSOFT_ASSERT(NULL != str2);

            return -1;
        }
        else if(NULL == str2)
        {
            STLSOFT_ASSERT(NULL != str1);

            return +1;
        }
        else
        {
            return ::strcmp(str1, str2);
        }
    }

    static int xtests_stricmp_(char const* str1, char const* str2)
    {
        if(str1 == str2)
        {
            return 0;
        }
        else if(NULL == str1)
        {
            STLSOFT_ASSERT(NULL != str2);

            return -1;
        }
        else if(NULL == str2)
        {
            STLSOFT_ASSERT(NULL != str1);

            return +1;
        }
        else
        {
            const size_t    len1    =   ::strlen(str1);
            const size_t    len2    =   ::strlen(str2);
            char_buffer_t_  str1_(1 + len1);
            char_buffer_t_  str2_(1 + len2);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
            if( str1_.empty() ||
                str2_.empty())
            {
                return 1;
            }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

#ifdef XTESTS_USING_SAFE_STR_FUNCTIONS
            ::strcpy_s(&str1_[0], str1_.size(), str1);
            ::strcpy_s(&str2_[0], str2_.size(), str2);
#else /* ? XTESTS_USING_SAFE_STR_FUNCTIONS */
            ::strcpy(&str1_[0], str1);
            ::strcpy(&str2_[0], str2);
#endif /* XTESTS_USING_SAFE_STR_FUNCTIONS */

            ::stlsoft::make_upper(str1_);
            ::stlsoft::make_upper(str2_);

            return ::strcmp(str1_.data(), str2_.data());
        }
    }

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
} // anonymous namespace
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////
 * API Implementation
 */

XTESTS_CALL(int) xtests_startRunner(char const* name, int verbosity, xTests_Reporter_t* reporter, void* reporterParam)
{
    STLSOFT_MESSAGE_ASSERT("Runner already initialised in this process!", NULL == s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        s_runner = new RunnerInfo(name, verbosity, reporter, reporterParam);

        if(NULL == s_runner)
        {
            return -2;
        }
        else
        {
            s_runner->PrintStart();

            return 0;
        }

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc&)
    {
        return -2;
    }
    catch(std::exception&)
    {
        return -1;
    }
    catch(...)
    {
        return -3;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(int) xtests_endRunner(int *retCode)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

    /* Bit of Null Object Pattern (Variable Variant) to assist */

    int retCode_ = EXIT_SUCCESS;

    if(NULL == retCode)
    {
        retCode = &retCode_;
    }

    if( 0 != s_runner->NumberOfFailedTestCases() &&
        EXIT_SUCCESS == *retCode)
    {
        *retCode = EXIT_FAILURE;
    }

    delete s_runner;

    s_runner = NULL;

    return *retCode;
}

XTESTS_CALL(void) xtests_printRunnerResults()
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

    s_runner->PrintResults();
}

XTESTS_CALL(void) xtests_abend(char const* message)
{
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        if(NULL != s_runner)
        {
            s_runner->onAbend(message);
        }

        xtests_output_(xtestsAlert, message);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(...)
    {}
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

    ::exit(EXIT_FAILURE);
}




XTESTS_CALL(int) xtests_beginTestCase(char const* name, char const* description)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->BeginCase(name, description);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot create test due to memory exhaustion\n");

        return 1;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot create test '";
        msg += name;
        msg += "': ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 1;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(int) xtests_endTestCase(char const* name)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->EndCase(name);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot end test due to memory exhaustion\n");

        return 1;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot end test '";
        msg += name;
        msg += "': ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 1;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}


XTESTS_CALL(int) xtests_testPassed()
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        s_runner->RegisterSuccessfulCondition();

        return 1;

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(void) xtests_caseExcepted(char const* exceptionType, char const* exceptionMessage)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

    // Remove leading "class "
    if(exceptionType == ::strstr(exceptionType, "class "))
    {
        exceptionType += 6;
    }

    s_runner->CaseExcepted(exceptionType, exceptionMessage);
}

XTESTS_CALL(void) xtests_caseExceptionExpected(char const* exceptionType)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

    s_runner->CaseExceptionExpected(exceptionType);
}

XTESTS_CALL(int) xtests_testFailed( char const* file
                                ,   int         line
                                ,   char const* function
                                ,   char const* expr)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition(file, line, function, expr);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(int) xtests_floatingPointClose(double expected, double actual)
{
    if(expected == actual)
    {
        return 1;
    }

    if( actual * s_fpApproximateFactor >= expected &&
        expected * s_fpApproximateFactor >= actual)
    {
        return 1;
    }

    return 0;
}

XTESTS_CALL(double) xtests_setFloatingPointCloseFactor(double factor, double *old /* = NULL */)
{
    STLSOFT_ASSERT(0.0 != factor);

    double  old_ = s_fpApproximateFactor;

    s_fpApproximateFactor = factor;

    if(NULL != old)
    {
        *old = old_;
    }

    return old_;
}



XTESTS_CALL(int) xtests_testFailed_int( char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   int                 expected
                                    ,   int                 actual
                                    ,   xtests_comparison_t comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition_long(file, line, function, static_cast<long>(expected), static_cast<long>(actual), comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(int) xtests_testFailed_long(char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   long                expected
                                    ,   long                actual
                                    ,   xtests_comparison_t comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition_long(file, line, function, expected, actual, comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

XTESTS_CALL(int) xtests_testFailed_ulong(char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   unsigned long       expected
                                    ,   unsigned long       actual
                                    ,   xtests_comparison_t comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition_ulong(file, line, function, expected, actual, comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
XTESTS_CALL(int) xtests_testFailed_longlong(char const*                     file
                                        ,   int                             line
                                        ,   char const*                     function
                                        ,   stlsoft_ns_qual(ss_sint64_t)    expected
                                        ,   stlsoft_ns_qual(ss_sint64_t)    actual
                                        ,   xtests_comparison_t             comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition_longlong(file, line, function, expected, actual, comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
XTESTS_CALL(int) xtests_testFailed_double(  char const*                     file
                                        ,   int                             line
                                        ,   char const*                     function
                                        ,   double                          expected
                                        ,   double                          actual
                                        ,   xtests_comparison_t             comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->RegisterFailedCondition_double(file, line, function, expected, actual, comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}


XTESTS_CALL(int) xtests_testMultibyteStrings(   char const*         file
                                            ,   int                 line
                                            ,   char const*         function
                                            ,   char const*         expected
                                            ,   char const*         actual
                                            ,   xtests_comparison_t comp)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->TestMultibyteStrings(file, line, function, expected, actual, comp);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}



XTESTS_CALL(int) xtests_writeFailMessage(char const* file
                                    ,   int         line
                                    ,   char const* function
                                    ,   char const* message
                                    ,   char const* qualifyingInformation)
{
    STLSOFT_MESSAGE_ASSERT("Runner not initialised in this process!", NULL != s_runner);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        return s_runner->WriteFailMessage(file, line, function, message, qualifyingInformation);

#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc& /* x */)
    {
        xtests_output_(xtestsCritical, "cannot update test due to memory exhaustion\n");

        return 0;
    }
    catch(std::exception& x)
    {
        std::string msg;

        msg.reserve(200);

        msg += "Cannot update test: ";
        msg += x.what();
        msg += "\n";

        xtests_output_(xtestsCritical, msg.c_str());

        return 0;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
}

/* /////////////////////////////////////////////////////////////////////////
 * Helper functions
 */

XTESTS_CALL(int) xtests_commandLine_parseVerbosity(int argc, char** argv, int *verbosity)
{
    STLSOFT_ASSERT(NULL != verbosity);

    *verbosity = xtestsVerbositySummaryOnSuccess;

    static const char   s_verb[]    =   "--verbosity=";
    static const size_t s_cchVerb   =   STLSOFT_NUM_ELEMENTS(s_verb) - 1;

    { for(int i = 1; i < argc; ++i)
    {
        STLSOFT_ASSERT(NULL != argv[i]);

        if(argv[i] == ::strstr(argv[i], s_verb))
        {
            *verbosity = ::atoi(argv[i] + s_cchVerb);

            return i;
        }
    }}

    return 0;
}

/* /////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
namespace
{
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

/* static */ xTests_Reporter_t* RunnerInfo::get_reporter_(xTests_Reporter_t* reporter)
{
#ifdef STLSOFT_COMPILER_IS_BORLAND
# define xtestsVerbositySilent              ::xtests::c::xtestsVerbositySilent
# define xtestsVerbositySummaryOnError      ::xtests::c::xtestsVerbositySummaryOnError
# define xtestsVerbosityVerboseOnError      ::xtests::c::xtestsVerbosityVerboseOnError
# define xtestsVerbositySummaryOnSuccess    ::xtests::c::xtestsVerbositySummaryOnSuccess
# define xtestsVerbosityVerboseOnSuccess    ::xtests::c::xtestsVerbosityVerboseOnSuccess
#endif /* STLSOFT_COMPILER_IS_BORLAND */

    if(NULL == reporter)
    {
        // NOTE: Even though this is thread-unsafe in that there's a race
        // condition for the initialisation of the s_reporter instance (and
        // its vtable), there is no problem because multiple initialisation
        // will be benign.

        class fprintf_reporter
            : public xTests_Reporter_t
        {
        private: // Overrides
            virtual void onStartRunner(void* /* reporterParam */, char const* name, int verbosity)
            {
                switch(verbosity)
                {
                    case    xtestsVerbositySilent:              // Don't say anything at any point
                    case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                    case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                        break;
                    case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                    case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
                    default:
                        ::fprintf(stdout, "Test runner '%s' starting:\n", name);
                        break;
                }
            }

            virtual void onBeginTestCase(void* /* reporterParam */, char const* /* name */, char const* /* desc */, int /* verbosity */)
            {
            }

            virtual void onTestPassed(void* /* reporterParam */, int /* verbosity */)
            {
            }

            virtual void onTestFailed(void* /* reporterParam */, char const* file, int line, char const* function, char const* expr, int verbosity)
            {
                static const char*  s_fmts[] = 
                {
                        "%s(%d): test condition \"%s\" failed\n"
                    ,   "%s(%d): test condition \"%s\" failed in function %s\n"
                };
                char const*         fmt = s_fmts[NULL != function];

                switch(verbosity)
                {
                    case    xtestsVerbositySilent:              // Don't say anything at any point
                        break;
                    case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                        break;
                    case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                        break;
                    case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                        break;
                    case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
                    default:
                        break;
                }

                ::fprintf(stdout, fmt, file, line, expr, function);
            }

            virtual void onCaseExcepted(void* /* reporterParam */, char const* caseName, char const* exceptionType, char const* exceptionMessage, int verbosity)
            {
                int level = 0;

                switch(verbosity)
                {
                    case    xtestsVerbositySilent:              // Don't say anything at any point
                        break;
                    case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                        level = 1;
                        break;
                    case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                    case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                    case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
                    default:
                        level = 2;
                        break;
                }

                STLSOFT_ASSERT(level >= 0 && level < 3);

                static char const*  s_fmts[] = 
                {
                        ""
                    ,   "\t%s: UX '%s' rx; msg='%s'\n"
                    ,   "\tTest case '%s': received unexpected exception of type '%s', with message '%s'\n"
                };
                char const*         fmt = s_fmts[level];

                ::fprintf(  stdout
                        ,   fmt
                        ,   caseName
                        ,   exceptionType
                        ,   exceptionMessage);
            }

            virtual void onCaseExceptionExpected(void* /* reporterParam */, char const* caseName, char const* exceptionType, int verbosity)
            {
                int level = 0;

                switch(verbosity)
                {
                    case    xtestsVerbositySilent:              // Don't say anything at any point
                        break;
                    case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                        level = 1;
                        break;
                    case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                    case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                    case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
                    default:
                        level = 2;
                        break;
                }

                STLSOFT_ASSERT(level >= 0 && level < 3);

                static char const*  s_fmts[] = 
                {
                        ""
                    ,   "\t%s: EX '%s' not rx\n"
                    ,   "\tTest case '%s': expected exception of type '%s' was not received\n"
                };
                char const*         fmt = s_fmts[level];

                ::fprintf(  stdout
                        ,   fmt
                        ,   caseName
                        ,   exceptionType
                        );
            }

            virtual void onEndTestCase(void* /* reporterParam */, char const* /* name */, xTests_runner_results_t const* /* results */, int /* verbosity */)
            {
            }

            virtual void onPrintRunnerResults(void* /* reporterParam */, xTests_runner_results_t const* /* results */, int /* verbosity */)
            {
            }

            virtual void onAbend(void* /* reporterParam */, char const* /* message */, int /* verbosity */)
            {
            }

            virtual void onDefect(void* /* reporterParam */, char const* /* message */, char const* /* qualifier */, int /* verbosity */)
            {
                ;
            }

            virtual void onEndRunner(void* /* reporterParam */, char const* /* name */, int /* verbosity */)
            {
            }
        };

        static fprintf_reporter s_reporter;

        reporter = &s_reporter;
    }

    return reporter;
}


RunnerInfo::RunnerInfo(char const* name, int verbosity, xTests_Reporter_t* reporter, void* reporterParam)
    : m_reporter(get_reporter_(reporter))
    , m_reporterParam(reporterParam)
    , m_name(name)
    , m_verbosity(verbosity)
    , m_tests()
    , m_totalConditions(0)
    , m_failedConditions(0)
    , m_unexpectedExceptions(0)
    , m_expectedExceptions(0)
    , m_currentCase(m_tests.end())
{
    STLSOFT_ASSERT(NULL != m_reporter);

    STLSOFT_MESSAGE_ASSERT("Invalid verbosity; must be >= 0", verbosity >= 0);
}

int RunnerInfo::BeginCase(char const* name, char const* description)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() != m_tests.find(name))
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("test case already exists");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "test case already exists", name);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else if(m_tests.end() != m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("a test case is already in progress; call XTESTS_CASE_END() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "a test case is already in progress; call XTESTS_CASE_END() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        m_currentCase = m_tests.insert(std::make_pair(name, TestInfo(name, description))).first;
    }

    return 0;
}

int RunnerInfo::EndCase(char const* /* name */)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo const& testInfo = (*m_currentCase).second;
        int             level = 0;

        switch(m_verbosity)
        {
            case    xtestsVerbositySilent:              // Don't say anything at any point
                level = 0;
                break;
            case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                if(!testInfo.haveAllTestsPassed())
                {
                    level = 1;
                }
                break;
            case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                level = 1;
                // Fall through
            case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                if(!testInfo.haveAllTestsPassed())
                {
                    level = 2;
                }
                break;
            case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
            default:
                level = 2;
                break;
        }
        STLSOFT_ASSERT(level >= 0 && level < 3);

        static char const*  s_fmts[] = 
        {
                ""
            ,   "\t%s: %u / %u / %u / %u / %u; result=%s\n"
            ,   "\tTest case '%s': %u assertions; %u succeeded; %u failed; %u unexpected exceptions; %u missing expected exceptions; result=%s\n"
        };
        char const*         fmt = s_fmts[level];

        ::fprintf(  stdout
                ,   fmt
                ,   testInfo.name.c_str()
                ,   testInfo.totalConditions
                ,   testInfo.totalConditions - testInfo.failedConditions
                ,   testInfo.failedConditions
                ,   testInfo.unexpectedExceptions
                ,   testInfo.expectedExceptions
                ,   testInfo.haveAllTestsPassed() ? "SUCCESS" : "FAILURE"
                );

        m_totalConditions       +=  testInfo.totalConditions;
        m_failedConditions      +=  testInfo.failedConditions;
        m_unexpectedExceptions  +=  testInfo.unexpectedExceptions;
        m_expectedExceptions    +=  testInfo.expectedExceptions;

        m_currentCase = m_tests.end();

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

int RunnerInfo::RegisterSuccessfulCondition()
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.totalConditions;

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

int RunnerInfo::RegisterFailedCondition(char const* file
                                    ,   int         line
                                    ,   char const* function
                                    ,   char const* expr)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.failedConditions;
        ++testInfo.totalConditions;

        m_reporter->onTestFailed(m_reporterParam, file, line, function, expr, m_verbosity);

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

int RunnerInfo::RegisterFailedCondition_long(   char const*         file
                                            ,   int                 line
                                            ,   char const*         function
                                            ,   long                expected
                                            ,   long                actual
                                            ,   xtests_comparison_t comp)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.failedConditions;
        ++testInfo.totalConditions;

        static char const*  s_fmts[] =
        {
                "%s(%d): test condition failed: actual value %ld should equal the expected value %ld%s%s\n"
            ,   "%s(%d): test condition failed: actual value %ld should not equal expected value %ld\n%s%s"
            ,   "%s(%d): test condition failed: actual value %ld should be greater than expected value %ld%s%s\n"
            ,   "%s(%d): test condition failed: actual value %ld should be less than expected value %ld%s%s\n"
            ,   "%s(%d): test condition failed: actual value %ld should be greater than or equal to the expected value %ld%s%s\n"
            ,   "%s(%d): test condition failed: actual value %ld should be less than or equal to the expected value %ld%s%s\n"
            ,   "%s(%d): test condition failed: actual value %ld should equal the expected value %ld%s%s\n"
        };
        STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == xtestsComparison_max_enumerator);
        char const*         fmt = s_fmts[comp];

        switch(m_verbosity)
        {
            case    xtestsVerbositySilent:              // Don't say anything at any point
                break;
            case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                break;
            case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                break;
            case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                break;
            case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
            default:
                break;
        }

        ::fprintf(stdout, fmt, file, line, actual, expected, (NULL != function) ? " in function " : "", stlsoft_ns_qual(c_str_ptr)(function));

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

int RunnerInfo::RegisterFailedCondition_ulong(  char const*         file
                                            ,   int                 line
                                            ,   char const*         function
                                            ,   unsigned long       expected
                                            ,   unsigned long       actual
                                            ,   xtests_comparison_t comp)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.failedConditions;
        ++testInfo.totalConditions;

        static char const*  s_fmts[] =
        {
                "%s(%d): test condition failed: actual value %lu should equal the expected value %lu%s%s\n"
            ,   "%s(%d): test condition failed: actual value %lu should not equal expected value %lu\n%s%s"
            ,   "%s(%d): test condition failed: actual value %lu should be greater than expected value %lu%s%s\n"
            ,   "%s(%d): test condition failed: actual value %lu should be less than expected value %lu%s%s\n"
            ,   "%s(%d): test condition failed: actual value %lu should be greater than or equal to the expected value %lu%s%s\n"
            ,   "%s(%d): test condition failed: actual value %lu should be less than or equal to the expected value %lu%s%s\n"
            ,   "%s(%d): test condition failed: actual value %lu should equal the expected value %lu%s%s\n"
        };
        STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == xtestsComparison_max_enumerator);
        char const*         fmt = s_fmts[comp];

        switch(m_verbosity)
        {
            case    xtestsVerbositySilent:              // Don't say anything at any point
                break;
            case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                break;
            case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                break;
            case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                break;
            case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
            default:
                break;
        }

        ::fprintf(stdout, fmt, file, line, actual, expected, (NULL != function) ? " in function " : "", stlsoft_ns_qual(c_str_ptr)(function));

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
int RunnerInfo::RegisterFailedCondition_longlong(char const*       file
                                            ,   int                 line
                                            ,   char const*         function
                                            ,   sint64_t            expected
                                            ,   sint64_t            actual
                                            ,   xtests_comparison_t comp)
{
    STLSOFT_ASSERT(NULL != m_reporter);

# if defined(STLSOFT_COMPILER_IS_BORLAND)
#  define static
# endif /* compiler */

    // Note: The following code has a race condition, but it is entirely benign, so does not matter
    static char const*  s_fmtBases[] =
    {
            "%%s(%%d): test condition failed: actual value %s should equal the expected value %s%%s%%s\n"
        ,   "%%s(%%d): test condition failed: actual value %s should not equal expected value %ld\n%s%s"
        ,   "%%s(%%d): test condition failed: actual value %s should be greater than expected value %s%%s%%s\n"
        ,   "%%s(%%d): test condition failed: actual value %s should be less than expected value %s%%s%%s\n"
        ,   "%%s(%%d): test condition failed: actual value %s should be greater than or equal to the expected value %s%%s%%s\n"
        ,   "%%s(%%d): test condition failed: actual value %s should be less than or equal to the expected value %s%%s%%s\n"
        ,   "%%s(%%d): test condition failed: actual value %s should equal the expected value %s%%s%%s\n"
    };
    STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmtBases) == xtestsComparison_max_enumerator);
    static char         s_fmts[7][101];
    STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == xtestsComparison_max_enumerator);
    STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == STLSOFT_NUM_ELEMENTS_(s_fmtBases));
    static char const*  s_fmt64 =   stlsoft::integral_printf_traits<stlsoft::sint64_t>::format_a();
    static const int    s_len   =   0
#ifdef XTESTS_USING_SAFE_STR_FUNCTIONS
                                +   ::sprintf_s(&s_fmts[0][0], STLSOFT_NUM_ELEMENTS_(s_fmts[0]), s_fmtBases[0], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[1][0], STLSOFT_NUM_ELEMENTS_(s_fmts[1]), s_fmtBases[1], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[2][0], STLSOFT_NUM_ELEMENTS_(s_fmts[2]), s_fmtBases[2], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[3][0], STLSOFT_NUM_ELEMENTS_(s_fmts[3]), s_fmtBases[3], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[4][0], STLSOFT_NUM_ELEMENTS_(s_fmts[4]), s_fmtBases[4], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[5][0], STLSOFT_NUM_ELEMENTS_(s_fmts[5]), s_fmtBases[5], s_fmt64, s_fmt64)
                                +   ::sprintf_s(&s_fmts[6][0], STLSOFT_NUM_ELEMENTS_(s_fmts[6]), s_fmtBases[6], s_fmt64, s_fmt64)
#else /* ? XTESTS_USING_SAFE_STR_FUNCTIONS */
                                +   ::sprintf(&s_fmts[0][0], s_fmtBases[0], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[1][0], s_fmtBases[1], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[2][0], s_fmtBases[2], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[3][0], s_fmtBases[3], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[4][0], s_fmtBases[4], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[5][0], s_fmtBases[5], s_fmt64, s_fmt64)
                                +   ::sprintf(&s_fmts[6][0], s_fmtBases[6], s_fmt64, s_fmt64)
#endif /* XTESTS_USING_SAFE_STR_FUNCTIONS */
                                +   0;

# if defined(STLSOFT_COMPILER_IS_BORLAND)
#  undef static
# endif /* compiler */

    char const*         fmt = s_fmts[comp];

    switch(m_verbosity)
    {
        case    xtestsVerbositySilent:              // Don't say anything at any point
            break;
        case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
            break;
        case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
            break;
        case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
            break;
        case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
        default:
            break;
    }

    ::fprintf(stdout, fmt, file, line, actual, expected, (NULL != function) ? " in function " : "", stlsoft_ns_qual(c_str_ptr)(function));

    STLSOFT_SUPPRESS_UNUSED(s_len);

    return 0;
}
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */

int RunnerInfo::RegisterFailedCondition_double( char const*         file
                                            ,   int                 line
                                            ,   char const*         function
                                            ,   double              expected
                                            ,   double              actual
                                            ,   xtests_comparison_t comp)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.failedConditions;
        ++testInfo.totalConditions;

        static char const*  s_fmts[] =
        {
                "%s(%d): test condition failed: actual value %G should equal the expected value %G%s%s\n"
            ,   "%s(%d): test condition failed: actual value %G should not equal expected value %G\n%s%s"
            ,   "%s(%d): test condition failed: actual value %G should be greater than expected value %G%s%s\n"
            ,   "%s(%d): test condition failed: actual value %G should be less than expected value %G%s%s\n"
            ,   "%s(%d): test condition failed: actual value %G should be greater than or equal to the expected value %G%s%s\n"
            ,   "%s(%d): test condition failed: actual value %G should be less than or equal to the expected value %G%s%s\n"
            ,   "%s(%d): test condition failed: actual value %G should be approximately equal to the expected value %G%s%s\n"
        };
        STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == xtestsComparison_max_enumerator);
        char const*         fmt = s_fmts[comp];

        switch(m_verbosity)
        {
            case    xtestsVerbositySilent:              // Don't say anything at any point
                break;
            case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                break;
            case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                break;
            case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                break;
            case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
            default:
                break;
        }

        ::fprintf(stdout, fmt, file, line, actual, expected, (NULL != function) ? " in function " : "", stlsoft_ns_qual(c_str_ptr)(function));

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}


int RunnerInfo::TestMultibyteStrings(   char const*         file
                                    ,   int                 line
                                    ,   char const*         function
                                    ,   char const*         expected
                                    ,   char const*         actual
                                    ,   xtests_comparison_t comp)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.totalConditions;

        // Now do test, taking into account NULL pointers
        int comparisonSucceeded = false;

        switch(comp)
        {
            case    xtestsComparisonEqual:
                if(0 == xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonApproxEqual:
                if(0 == xtests_stricmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonNotEqual:
                if(0 != xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonGreaterThan:
                if(0 > xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonLessThan:
                if(0 < xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonGreaterThanOrEqual:
                if(0 >= xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparisonLessThanOrEqual:
                if(0 <= xtests_strcmp_(expected, actual))
                {
                    comparisonSucceeded = true;
                }
                break;
            case    xtestsComparison_max_enumerator:
                xtests_abend("invalid test comparison type: test framework may be out of date!");
                break;
        }

        if(comparisonSucceeded)
        {
            return 1;
        }
        else
        {
            ++testInfo.failedConditions;

            // Eliminate NULL pointers
            expected = stlsoft::c_str_ptr_a(expected);
            actual = stlsoft::c_str_ptr_a(actual);

            static char const*  s_fmts[] =
            {
                    "%s(%d): test condition failed: actual value '%s' should equal the expected value '%s'%s%s\n"
                ,   "%s(%d): test condition failed: actual value '%s' should not equal expected value '%s'\n%s%s"
                ,   "%s(%d): test condition failed: actual value '%s' should be greater than expected value '%s'%s%s\n"
                ,   "%s(%d): test condition failed: actual value '%s' should be less than expected value '%s'%s%s\n"
                ,   "%s(%d): test condition failed: actual value '%s' should be greater than or equal to the expected value '%s'%s%s\n"
                ,   "%s(%d): test condition failed: actual value '%s' should be less than or equal to the expected value '%s'%s%s\n"
                ,   "%s(%d): test condition failed: actual value '%s' should be approximately equal to the expected value '%s'%s%s\n"
            };
            STLSOFT_STATIC_ASSERT(STLSOFT_NUM_ELEMENTS_(s_fmts) == xtestsComparison_max_enumerator);
            char const*         fmt = s_fmts[comp];

            switch(m_verbosity)
            {
                case    xtestsVerbositySilent:              // Don't say anything at any point
                    break;
                case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                    break;
                case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                    break;
                case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                    break;
                case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
                default:
                    break;
            }

            ::fprintf(stdout, fmt, file, line, actual, expected, (NULL != function) ? " in function " : "", stlsoft_ns_qual(c_str_ptr)(function));

            return 0;
        }
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}


int RunnerInfo::WriteFailMessage(   char const* file
                                ,   int         line
                                ,   char const* function
                                ,   char const* message
                                ,   char const* qualifyingInformation)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);

        return -1;
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo& testInfo = (*m_currentCase).second;

        ++testInfo.failedConditions;
        ++testInfo.totalConditions;

        static const char  s_fmt[] = "%s(%d): %s%s%s%s%s\n";

        switch(m_verbosity)
        {
            case    xtestsVerbositySilent:              // Don't say anything at any point
                break;
            case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
                break;
            case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
                break;
            case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
                break;
            case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
            default:
                break;
        }

        ::fprintf(stdout, s_fmt, file, line, message, (NULL != function) ? " in function " : "", (NULL != function) ? function : "", (NULL != qualifyingInformation) ? ": " : "", (NULL != qualifyingInformation) ? qualifyingInformation : "");

        return 0;
    }

#if defined(STLSOFT_COMPILER_IS_DMC)
    return -1;
#endif /* compiler */
}

void RunnerInfo::CaseExcepted(  char const* exceptionType
                            ,   char const* exceptionMessage)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo&   testInfo = (*m_currentCase).second;

        ++testInfo.unexpectedExceptions;

        m_reporter->onCaseExcepted(m_reporterParam, testInfo.name.c_str(), exceptionType, exceptionMessage, m_verbosity);
    }
}

void RunnerInfo::CaseExceptionExpected(char const* exceptionType)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if(m_tests.end() == m_currentCase)
    {
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
        throw std::runtime_error("not in a test case; call XTESTS_CASE_BEGIN() ?");
#else /* ? STLSOFT_CF_EXCEPTION_SUPPORT */
        m_reporter->onDefect(m_reporterParam, "not in a test case; call XTESTS_CASE_BEGIN() ?", NULL);
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */
    }
    else
    {
        TestInfo&   testInfo = (*m_currentCase).second;

        ++testInfo.expectedExceptions;

        m_reporter->onCaseExceptionExpected(m_reporterParam, testInfo.name.c_str(), exceptionType, m_verbosity);
    }
}

void RunnerInfo::PrintStart()
{
    STLSOFT_ASSERT(NULL != m_reporter);

    m_reporter->onStartRunner(m_reporterParam, m_name.c_str(), m_verbosity);
}

void RunnerInfo::PrintResults()
{
    STLSOFT_ASSERT(NULL != m_reporter);

    int level = 0;

    switch(m_verbosity)
    {
        case    xtestsVerbositySilent:              // Don't say anything at any point
            level = 0;
            break;
        case    xtestsVerbositySummaryOnError:      // Show a summary, but only on error
            if(NumberOfFailedTestCases())
            {
                level = 1;
            }
            break;
        case    xtestsVerbositySummaryOnSuccess:    // Give full information on error, otherwise just summary
            // Fall through
        case    xtestsVerbosityVerboseOnError:      // Give full information, but only on error */
            level = 1;
            if(NumberOfFailedTestCases())
            {
                level = 2;
            }
            break;
        case    xtestsVerbosityVerboseOnSuccess:    // Give full information at all times
        default:
            level = 2;
            break;
    }

    STLSOFT_ASSERT(level >= 0 && level < 3);

    static char const*  s_fmts[] = 
    {
            ""
        ,   "Test runner '%s' complete:"
            "\t%u / %u / %u / %u / %u / %u; result=%s\n"
            "\n"
        ,   "------------------------------------------------------------\n"
            "Test runner '%s' complete:\n"
            "\t%u test case(s)\n"
            "\t%u total assertion(s)\n"
            "\t%u total assertion(s) succeeded\n"
            "\t%u total assertion(s) failed\n"
            "\t%u total unexpected exception(s)\n"
            "\t%u total missing expected exception(s)\n"
            "\tresult=%s\n"
            "\n"
    };
    char const*         fmt = s_fmts[level];

    ::fprintf(  stdout
            ,   fmt
            ,   m_name.c_str()
            ,   static_cast<unsigned>(m_tests.size())
            ,   static_cast<unsigned>(m_totalConditions)
            ,   static_cast<unsigned>(m_totalConditions - m_failedConditions)
            ,   static_cast<unsigned>(m_failedConditions)
            ,   static_cast<unsigned>(m_unexpectedExceptions)
            ,   static_cast<unsigned>(m_expectedExceptions)
            ,   (   0u == m_failedConditions &&
                    0u ==  m_unexpectedExceptions &&
                    0u == m_expectedExceptions) ? "SUCCESS" : "FAILURE"
            );

}

void RunnerInfo::onAbend(char const* message)
{
    STLSOFT_ASSERT(NULL != m_reporter);

    m_reporter->onAbend(m_reporterParam, message, m_verbosity);
}

int RunnerInfo::NumberOfFailedTestCases() const
{
    STLSOFT_ASSERT(NULL != m_reporter);

    if( 0u == m_failedConditions &&
        0u == m_unexpectedExceptions &&
        0u == m_expectedExceptions)
    {
        return 0;
    }
    else
    {
        int numFailed = 0;

        { for(test_map_t::const_iterator b = m_tests.begin(); b != m_tests.end(); ++b)
        {
            TestInfo const& testInfo = (*b).second;

            if(!testInfo.haveAllTestsPassed())
            {
                ++numFailed;
            }
        }}

        return numFailed;
    }
}

#ifdef STLSOFT_CF_NAMESPACE_SUPPORT
} // anonymous namespace
#endif /* STLSOFT_CF_NAMESPACE_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _XTESTS_NO_NAMESPACE
} /* namespace c */
} /* namespace xtests */
#endif /* !_XTESTS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////// */
